#include <cstring>
#include <iostream>
#include "TarHandler.h"

#define ASCII_TO_NUMBER(num) ((num) - 48) //Converts an ascii digit to the corresponding number

TarHandler::TarHandler()
    : idle(true)
    , tar_index(0)
    , next_tar_header(0)
    , file_size(0)
    , read_byte_count(0)
    , processed_tar_blocks(0)
    , file_processed_bytes(0)
{
}

void TarHandler::tcp_in_handler(char* data, std::size_t size)
{
    if(this->idle) {
        this->idle = false;
        this->tar_index = 0;
        this->file_size = 0;
        this->next_tar_header;
        this->read_byte_count = 0;
        this->file_size = 0;
    }

    // process incoming buffer
    for(std::size_t i = 0; i < size; i++) {
        this->tar_block[this->tar_index] = data[i];
        this->read_byte_count++;
        if(this->tar_index++ == TAR_BLOCK_SIZE - 1) {
            this->tar_block_handler();
            this->tar_index = 0;
        }
    }
}

bool TarHandler::tar_block_handler()
{
    this->processed_tar_blocks++;
    if(this->next_tar_header == (this->read_byte_count) - TAR_BLOCK_SIZE) {
        // Tar header detected
        struct tar_file_header tar_header;
        memcpy(&tar_header, this->tar_block, TAR_BLOCK_SIZE);
        if(!check_checksum(&tar_header)) {
            fprintf(stderr, "checksum missmatch\r\n");
            return false;
        }
        this->file_size = decode_tar_octal(tar_header.file_size, sizeof(tar_header.file_size));
        this->file_processed_bytes = 0;
        printf("%s - size: %lu\r\n", tar_header.filename, file_size);
        this->next_tar_header = (file_size / TAR_BLOCK_SIZE + 1) * TAR_BLOCK_SIZE + this->processed_tar_blocks * TAR_BLOCK_SIZE;
    } else {
        std::size_t i = 0;
        if((this->file_size - this->file_processed_bytes) >= TAR_BLOCK_SIZE) {
            for(i = 0; i < TAR_BLOCK_SIZE; i++) {
                printf("%c", this->tar_block[i]);
            }
        }
        else {
            for(i = 0; i < (this->file_size - this->file_processed_bytes); i++) {
                printf("%c", this->tar_block[i]);
            }
        }
        this->file_processed_bytes += i;
        printf("processed bytes: %lu\r\n", this->file_processed_bytes);
    }
    return true;
}

uint64_t TarHandler::decode_tar_octal(char* data, std::size_t size)
{
    unsigned char *current_ptr = (unsigned char *) data + size;
    uint64_t sum = 0;
    uint64_t current_multiplier = 1;

    unsigned char* check_ptr = current_ptr; //This is used to check where the last NUL/space char is
    for(; check_ptr >= (unsigned char *) data; --check_ptr) {
        if((*check_ptr) == 0 || (*check_ptr) == ' ') {
            current_ptr = check_ptr - 1;
        }
    }
    for(; current_ptr >= (unsigned char *) data; --current_ptr) {
        sum += ASCII_TO_NUMBER(*current_ptr) * current_multiplier;
        current_multiplier *= 8;
    }
    return sum;
}

bool TarHandler::check_checksum(struct tar_file_header *tar_header)
{
    char original_checksum[8];
    memcpy(original_checksum, tar_header->checksum, 8);
    memset(tar_header->checksum, ' ', 8);

    int64_t unsigned_sum = 0;
    int64_t signed_sum = 0;
    unsigned char *uc_tar = (unsigned char *)tar_header;
    signed char *sc_tar = (signed char *)tar_header;
    for(int i = 0; i < TAR_BLOCK_SIZE; i++) {
        unsigned_sum += uc_tar[i];
        signed_sum += sc_tar[i];
    }
    //Copy back the checksum
    memcpy(tar_header->checksum, original_checksum, 8);
    //Decode the original checksum
    uint64_t reference_checksum = decode_tar_octal(original_checksum, sizeof(original_checksum));
    return (reference_checksum == unsigned_sum || reference_checksum == signed_sum);
}
